/*
 VARNA is a tool for the automated drawing, visualization and annotation of the secondary structure of RNA, designed as a companion software for web servers and databases.
 Copyright (C) 2008  Kevin Darty, Alain Denise and Yann Ponty.
 electronic mail : Yann.Ponty@lri.fr
 paper mail : LRI, bat 490 Universit Paris-Sud 91405 Orsay Cedex France

 This file is part of VARNA version 3.1.
 VARNA version 3.1 is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

 VARNA version 3.1 is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along with VARNA version 3.1.
 If not, see http://www.gnu.org/licenses.
 */
package fr.orsay.lri.varna.controlers;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.Vector;

import fr.orsay.lri.varna.VARNAPanel;
import fr.orsay.lri.varna.exceptions.MappingException;
import fr.orsay.lri.varna.models.VARNAConfig;
import fr.orsay.lri.varna.models.rna.Mapping;
import fr.orsay.lri.varna.models.rna.ModeleBase;
import fr.orsay.lri.varna.models.rna.RNA;

public class ControleurInterpolator extends Thread {

	VARNAPanel _vpn;
	private int _numSteps = 25;
	private long _timeDelay = 15;
	private boolean _running = false;
	Targets _d = new Targets();

	public ControleurInterpolator(VARNAPanel vpn) {
		_vpn = vpn;
	}

	public synchronized void addTarget(RNA target, Mapping mapping) {
		addTarget(target, null, mapping);
	}

	
	public synchronized void addTarget(RNA target, VARNAConfig conf, Mapping mapping) {
		_d.add(new TargetsHolder(target,conf,mapping));
	}

	public synchronized void addTarget(RNA target) {
		addTarget(target, null, Mapping.DefaultOutermostMapping(_vpn.getRNA()
				.get_listeBases().size(), target.get_listeBases().size()));
	}
	
	public boolean isInterpolationInProgress()
	{
		return _running;
	}

	private Point2D.Double computeDestination(Point2D pli, Point2D pri,
			Point2D pi, int i, int n, Point2D plf, Point2D prf) {
		Point2D.Double plm = new Point2D.Double(
				(pli.getX() + plf.getX()) / 2.0,
				(pli.getY() + plf.getY()) / 2.0);
		Point2D.Double prm = new Point2D.Double(
				(pri.getX() + prf.getX()) / 2.0,
				(pri.getY() + prf.getY()) / 2.0);
		Point2D.Double pm = new Point2D.Double(((n - i) * plm.getX() + i
				* prm.getX())
				/ n, ((n - i) * plm.getY() + i * prm.getY()) / n);
		Point2D.Double v = new Point2D.Double(pm.getX() - pi.getX(), pm.getY()
				- pi.getY());
		Point2D.Double pf = new Point2D.Double(pi.getX() + 2.0 * v.getX(), pi
				.getY()
				+ 2.0 * v.getY());
		return pf;
	}

	private Vector<Vector<Integer>> clusterIndices(int numIndices,
			int[] mappedIndices) throws MappingException {
		int[] indices = new int[numIndices];
		for (int i = 0; i < numIndices; i++) {
			indices[i] = i;
		}
		return clusterIndices(indices, mappedIndices);
	}

	/**
	 * Builds and returns an interval array, alternating unmatched regions and
	 * matched indices. Namely, returns a vector of vector such that the vectors
	 * found at odd positions contain those indices that are NOT associated with
	 * any other base in the current mapping. On the other hand, vectors found
	 * at even positions contain only one mapped index. <br/>
	 * Ex: If indices=<code>[1,2,3,4,5]</code> and mappedIndices=
	 * <code>[2,5]</code> then the function will return
	 * <code>[[1],[2],[3,4],[5],[]]</code>.
	 * 
	 * @param indices
	 *            The total list of indices
	 * @param mappedIndices
	 *            Matched indices, should be a subset of <code>indices</code>
	 * @return A clustered array
	 * @throws MappingException
	 *             If one of the parameters is an empty array
	 */
	private Vector<Vector<Integer>> clusterIndices(int[] indices,
			int[] mappedIndices) throws MappingException {
		if ((mappedIndices.length == 0) || (indices.length == 0)) {
			throw new MappingException(
					"Mapping Error: Cannot cluster indices in an empty mapping");
		}
		Vector<Vector<Integer>> res = new Vector<Vector<Integer>>();

		Arrays.sort(indices);
		Arrays.sort(mappedIndices);
		int i, j = 0, k;
		Vector<Integer> tmp = new Vector<Integer>();
		for (i = 0; (i < indices.length) && (j < mappedIndices.length); i++) {
			if (indices[i] == mappedIndices[j]) {
				res.add(tmp);
				tmp = new Vector<Integer>();
				tmp.add(indices[i]);
				res.add(tmp);
				tmp = new Vector<Integer>();
				j++;
			} else {
				tmp.add(indices[i]);
			}
		}
		k = i;
		for (i = k; (i < indices.length); i++) {
			tmp.add(indices[i]);
		}
		res.add(tmp);
		return res;
	}
	
	

	public void run() {
		while (true)
		{
			TargetsHolder d = _d.get();
			_running = true;
			try{
			nextTarget(d.target,d.conf,d.mapping);
			}
			catch(Exception e)
			{
				System.err.println(e);
				e.printStackTrace();
			}
			_running = false;
		}
		
	}

	
	public void nextTarget(RNA _target, VARNAConfig _conf, Mapping _mapping)
	{
		try {
			RNA source = _vpn.getRNA();
			RNA current = source;
			

			ArrayList<ModeleBase> currBases = current.get_listeBases();
			ArrayList<ModeleBase> destBases = _target.get_listeBases();
			Vector<Vector<Integer>> intArrSource = new Vector<Vector<Integer>>();
			Vector<Vector<Integer>> intArrTarget = new Vector<Vector<Integer>>();
			try {
				// Building interval arrays
				intArrSource = clusterIndices(currBases.size(), _mapping
						.getSourceElems());
				intArrTarget = clusterIndices(destBases.size(), _mapping
						.getTargetElems());
			} catch (MappingException e) {
				e.printStackTrace();
			}

			// Duplicating source and target coordinates
			Point2D.Double[] initPosSource = new Point2D.Double[currBases
					.size()];
			Point2D.Double[] finalPosTarget = new Point2D.Double[destBases
					.size()];

			for (int i = 0; i < currBases.size(); i++) {
				Point2D tmp = currBases.get(i).getCoords();
				initPosSource[i] = new Point2D.Double(tmp.getX(), tmp.getY());
			}
			for (int i = 0; i < destBases.size(); i++) {
				Point2D tmp = destBases.get(i).getCoords();
				finalPosTarget[i] = new Point2D.Double(tmp.getX(), tmp.getY());
			}

			/**
			 * Assigning final (Source) and initial (Target) coordinates
			 */
			Point2D.Double[] finalPosSource = new Point2D.Double[initPosSource.length];
			Point2D.Double[] initPosTarget = new Point2D.Double[finalPosTarget.length];
			// Final position of source model
			for (int i = 0; i < finalPosSource.length; i++) {
				if (_mapping.getPartner(i) != Mapping.UNKNOWN) {
					Point2D dest;
					dest = finalPosTarget[_mapping.getPartner(i)];
					finalPosSource[i] = new Point2D.Double(dest.getX(), dest
							.getY());
				}
			}

			for (int i = 0; i < intArrSource.size(); i += 2) {
				int matchedNeighborLeft, matchedNeighborRight;
				if (i == 0) {
					matchedNeighborLeft = intArrSource.get(1).get(0);
					matchedNeighborRight = intArrSource.get(1).get(0);
				} else if (i == intArrSource.size() - 1) {
					matchedNeighborLeft = intArrSource.get(
							intArrSource.size() - 2).get(0);
					matchedNeighborRight = intArrSource.get(
							intArrSource.size() - 2).get(0);
				} else {
					matchedNeighborLeft = intArrSource.get(i - 1).get(0);
					matchedNeighborRight = intArrSource.get(i + 1).get(0);
				}
				Vector<Integer> v = intArrSource.get(i);
				for (int j = 0; j < v.size(); j++) {
					int index = v.get(j);
					finalPosSource[index] = computeDestination(
							initPosSource[matchedNeighborLeft],
							initPosSource[matchedNeighborRight],
							initPosSource[index], j + 1, v.size() + 1,
							finalPosSource[matchedNeighborLeft],
							finalPosSource[matchedNeighborRight]);
				}
			}
			for (int i = 0; i < initPosTarget.length; i++) {
				if (_mapping.getAncestor(i) != Mapping.UNKNOWN) {
					Point2D dest;
					dest = initPosSource[_mapping.getAncestor(i)];
					initPosTarget[i] = new Point2D.Double(dest.getX(), dest
							.getY());
				}
			}
			for (int i = 0; i < intArrTarget.size(); i += 2) {
				int matchedNeighborLeft, matchedNeighborRight;
				if (i == 0) {
					matchedNeighborLeft = intArrTarget.get(1).get(0);
					matchedNeighborRight = intArrTarget.get(1).get(0);
				} else if (i == intArrTarget.size() - 1) {
					matchedNeighborLeft = intArrTarget.get(
							intArrTarget.size() - 2).get(0);
					matchedNeighborRight = intArrTarget.get(
							intArrTarget.size() - 2).get(0);
				} else {
					matchedNeighborLeft = intArrTarget.get(i - 1).get(0);
					matchedNeighborRight = intArrTarget.get(i + 1).get(0);
				}
				Vector<Integer> v = intArrTarget.get(i);
				for (int j = 0; j < v.size(); j++) {
					int index = v.get(j);
					initPosTarget[index] = computeDestination(
							finalPosTarget[matchedNeighborLeft],
							finalPosTarget[matchedNeighborRight],
							finalPosTarget[index], j + 1, v.size() + 1,
							initPosTarget[matchedNeighborLeft],
							initPosTarget[matchedNeighborRight]);
				}
			}

			boolean firstHalf = true;
			for (int i = 0; i < _numSteps; i++) {
				if (i == _numSteps / 2) {
					_vpn.showRNA(_target);
					current = _target;
					currBases = current.get_listeBases();
					firstHalf = false;
					if (_conf!=null)
					{_vpn.setConfig(_conf);}
					
					for (int j = 0; j < initPosSource.length; j++) {
						source.setCoord(j, initPosSource[j]);
					}
				}
				for (int j = 0; j < currBases.size(); j++) {
					ModeleBase m = currBases.get(j);
					Point2D mpc, mnc;
					if (firstHalf) {
						mpc = initPosSource[j];
						mnc = finalPosSource[j];
					} else {
						mpc = initPosTarget[j];
						mnc = finalPosTarget[j];
					}
					m.setCoords(new Point2D.Double(((_numSteps - 1 - i)
							* mpc.getX() + (i) * mnc.getX())
							/ (_numSteps - 1), ((_numSteps - 1 - i)
							* mpc.getY() + i * mnc.getY())
							/ (_numSteps - 1)));
				}
				_vpn.repaint();
				sleep(_timeDelay);
			}
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		_vpn.showRNA(_target);
		_vpn.repaint();

	}

	private class TargetsHolder
	{
		public RNA target;
		public VARNAConfig conf;
		public Mapping mapping;
		public TargetsHolder(RNA t, VARNAConfig c, Mapping m)
		{
			target = t;
			conf = c;
			mapping = m;
		}
	}
	
	private class Targets
	{
		LinkedList<TargetsHolder> _d = new LinkedList<TargetsHolder>();
		public synchronized void add(TargetsHolder d)
		{
			_d.addLast(d);
			notify();
		}
		public synchronized TargetsHolder get()
		{
			while(_d.size()==0)
			{
				try {
					wait();
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
			TargetsHolder x = _d.getFirst();
			_d.removeFirst();
			return x;
		}
	}

}
